import typing as T
import random
import string
import re

ALIAS_PATTERN = r">\$(\w+)"
RE_CODE = "04730"
SEPARATOR = "__"


ListType = T.TypeVar("ListType")


def random_str(n: int, include_re_code: bool = False) -> str:
    pre_str = RE_CODE if include_re_code else ""
    return pre_str + "".join(random.choices(string.ascii_letters + string.digits, k=n))


def random_digits(n: int) -> str:
    return "".join(random.choices(string.digits, k=n))


def chunk_list(lst: list[ListType], chunk_size: int) -> list[list[ListType]]:
    return [lst[i : i + chunk_size] for i in range(0, len(lst), chunk_size)]


def flatten_list(lst: T.Iterable[T.Iterable[ListType]]) -> list[ListType]:
    return [j for sub in lst for j in sub]


def replace_str_with_vars(s: str, variables: dict[str, T.Any]) -> str:
    if not s:
        return ""
    for var_name, var_val in variables.items():
        s = re.sub(
            rf"(\$)({var_name})(\W*)", r"\1" + f"[({var_val})]" + r"\3", s
        ).replace(f"[({var_val})]", f"{var_val}")
    return s



import typing as T
import json
import re
import edgedb
from pydantic import BaseModel, PrivateAttr
from pydantic.main import ModelMetaclass
from edge_orm.node import Node, Insert, Patch, EdgeConfigBase
from edge_orm.logs import logger
from edge_orm.external import encoders
from edge_orm import helpers, execute, span
from . import enums, errors, utils
from .nested_resolvers import NestedResolvers
from devtools import debug
from .merging import merge_nested_resolver

NodeType = T.TypeVar("NodeType", bound=Node)
InsertType = T.TypeVar("InsertType", bound=Insert)
PatchType = T.TypeVar("PatchType", bound=Patch)

EdgeNodeType = T.TypeVar("EdgeNodeType", bound=T.Type[Node])
ThisResolverType = T.TypeVar("ThisResolverType", bound="Resolver")  # type: ignore

VARS = dict[str, T.Any]
CONVERSION_FUNC = T.Callable[[str], T.Any]
FILTER_FIELDS = ["_filter", "_limit", "_offset", "_order_by"]
RAW_RESP_ONE = dict[str, T.Any]
RAW_RESP_MANY = list[RAW_RESP_ONE]
RAW_RESPONSE = RAW_RESP_ONE | RAW_RESP_MANY


class Meta(ModelMetaclass):
    """adds property _node_config to resolver from _node_cls"""

    def __new__(mcs, name, bases, dct, **kwargs):  # type: ignore
        x = super().__new__(mcs, name, bases, dct, **kwargs)
        if "_node_cls" in dct:
            x._node_config: EdgeConfigBase = dct["_node_cls"].EdgeConfig  # type: ignore
        return x


class Resolver(BaseModel, T.Generic[NodeType, InsertType, PatchType], metaclass=Meta):
    _filter: str = PrivateAttr(None)
    _order_by: str = PrivateAttr(None)
    _limit: int = PrivateAttr(None)
    _offset: int = PrivateAttr(None)

    _query_variables: VARS = PrivateAttr(default_factory=dict)

    _fields_to_return: set[str] = PrivateAttr(default_factory=set)  # init this?
    _extra_fields: set[str] = PrivateAttr(default_factory=set)
    _extra_fields_conversion_funcs: dict[str, CONVERSION_FUNC] = PrivateAttr(
        default_factory=dict
    )

    _nested_resolvers: NestedResolvers = PrivateAttr(default_factory=NestedResolvers)

    _node_cls: T.ClassVar[T.Type[NodeType]]  # type: ignore
    _insert_cls: T.ClassVar[T.Type[InsertType]]  # type: ignore
    _patch_cls: T.ClassVar[T.Type[PatchType]]  # type: ignore

    _node_config: T.ClassVar[EdgeConfigBase]

    is_count: bool = False
    update_operation: enums.UpdateOperation | None = None
    _merged: bool = PrivateAttr(False)

    _edge_resolver_map: T.ClassVar[dict[str, T.Type["Resolver"]]]  # type: ignore

    def __init__(self, **data: T.Any) -> None:
        super().__init__(**data)
        if not self._fields_to_return:
            self._fields_to_return = (
                self.node_field_names()
                - self._node_config.appendix_properties
                - self._node_config.computed_properties
            )

    @property
    def model_name(self) -> str:
        return self._node_config.model_name

    @classmethod
    def node_field_names(cls: T.Type[ThisResolverType]) -> set[str]:
        return {field.alias for field in cls._node_cls.__fields__.values()}

    """RESOLVER BUILDING METHODS"""

    def set_update_operation(
        self: ThisResolverType,
        replace: bool = False,
        add: bool = False,
        remove: bool = False,
    ) -> ThisResolverType:
        if replace:
            self.update_operation = enums.UpdateOperation.REPLACE
        elif add:
            self.update_operation = enums.UpdateOperation.ADD
        elif remove:
            self.update_operation = enums.UpdateOperation.REMOVE
        else:
            raise errors.ResolverException("Invalid update operation given.")
        return self

    def add_query_variables(self, variables: VARS | None) -> None:
        """

        :param variables: a dictionary of query variables to smartly merge with _query_variables
        :return: None
        """
        # can vars be enums now? Or should I do this later?
        # also do conflict nested later
        if not variables:
            return
        for key, val in variables.items():
            if key in self._query_variables:
                if val is not self._query_variables[key]:
                    raise errors.ResolverException(
                        f"Variable {key}, {val=} is already used."
                    )
            self._query_variables[key] = val

    def filter(
        self: ThisResolverType,
        filter_str: str,
        variables: VARS | None = None,
        connector: enums.FilterConnector = enums.FilterConnector.AND,
    ) -> ThisResolverType:
        """

        :param filter_str: string to filter by, like .name = <str>$name
        :param variables: query variables from filter_str, like {"name": "Paul Graham"}
        :param connector: how to connect to an existing filter.
         If OR and (.name = <str>$name) was already set as the filter, the new filter would be:
         .name = <str>$name OR .slug = <str>$slug
        :return: the resolver
        """
        if self._filter and not connector:
            raise errors.ResolverException(
                f"Filter of {self._filter=} has already been provided so connector needed."
            )
        if not self._filter and connector is enums.FilterConnector.OR:
            raise errors.ResolverException(
                f"You cannot try to filter with OR while there is no existing filter."
            )
        self.add_query_variables(variables=variables)
        if self._filter:
            self._filter = f"{self._filter}{connector.value}{filter_str}"
        else:
            self._filter = filter_str
        return self

    def filter_str_from_field_name(self, field_name: str) -> str:
        cast = self._node_config.node_edgedb_conversion_map[field_name].cast
        return f".{field_name} = <{cast}>${field_name}"

    def _filter_by(
        self: ThisResolverType,
        connector: enums.FilterConnector = enums.FilterConnector.AND,
        **kwargs: T.Any,
    ) -> ThisResolverType:
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        if not kwargs:
            raise errors.ResolverException("Nothing to filter by.")
        conversion_map = self._node_config.node_edgedb_conversion_map
        filter_strs = []
        variables = {}
        for field_name, field_value in kwargs.items():
            cast = conversion_map[field_name].cast
            filter_strs.append(f".{field_name} = <{cast}>${field_name}")
            variables[field_name] = field_value
        filter_str = " AND ".join(filter_strs)
        return self.filter(
            filter_str=filter_str, variables=variables, connector=connector
        )

    def _filter_in(
        self: ThisResolverType,
        connector: enums.FilterConnector = enums.FilterConnector.AND,
        **kwargs: T.Any,
    ) -> ThisResolverType:
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        if not kwargs:
            raise errors.ResolverException("Nothing to filter by.")
        conversion_map = self._node_config.node_edgedb_conversion_map
        filter_strs = []
        variables = {}
        for field_name, value_lst in kwargs.items():
            cast = conversion_map[field_name].cast
            variable_name = field_name
            if cast.startswith("default::"):  # if an enum or other scalar
                s = f".{field_name} in <{cast}>array_unpack(<array<str>>${variable_name})"
            else:
                s = f".{field_name} in array_unpack(<array<{cast}>>${variable_name})"
            filter_strs.append(s)
            variables[variable_name] = value_lst
        filter_str = " AND ".join(filter_strs)
        return self.filter(
            filter_str=filter_str, variables=variables, connector=connector
        )

    def order_by(
        self: ThisResolverType,
        order_by_str: str,
        variables: VARS | None = None,
        then: bool = False,
    ) -> ThisResolverType:
        """

        :param order_by_str: string to order by, like .created_at ASC
        :param variables: query variables if used in the order_by_str
        :param then: allows you to append this order by if order by already exists
        :return: the resolver
        """
        if self._order_by and not then:
            raise errors.ResolverException(
                f"Order by of {self._order_by} has already been provided."
            )
        self.add_query_variables(variables)
        if self._order_by:
            self._order_by = f"{self._order_by} THEN {order_by_str}"
        else:
            self._order_by = order_by_str
        return self

    def offset(self: ThisResolverType, /, _: int | None) -> ThisResolverType:
        if self._offset is not None:
            raise errors.ResolverException(
                f"Offset of {self._offset} has already been provided."
            )
        self._offset = _
        return self

    def limit(self: ThisResolverType, /, _: int | None) -> ThisResolverType:
        if self._limit is not None:
            raise errors.ResolverException(
                f"Limit of {self._limit} has already been provided."
            )
        self._limit = _
        return self

    def include_fields(
        self: ThisResolverType, *fields_to_include: str
    ) -> ThisResolverType:
        self._fields_to_return.update(fields_to_include)
        return self

    def exclude_fields(
        self: ThisResolverType, *fields_to_exclude: str
    ) -> ThisResolverType:
        self._fields_to_return = self._fields_to_return - set(fields_to_exclude)
        return self

    def include_appendix_properties(self: ThisResolverType) -> ThisResolverType:
        self._fields_to_return.update(self._node_config.appendix_properties)
        return self

    def include_computed_properties(self: ThisResolverType) -> ThisResolverType:
        self._fields_to_return.update(self._node_config.computed_properties)
        return self

    def extra_field(
        self: ThisResolverType,
        field_name: str,
        expression: str,
        conversion_func: CONVERSION_FUNC | None = None,
    ) -> ThisResolverType:
        """extra fields do NOT take in variables"""
        extra_field_str = f"{field_name} := {expression}"
        self._extra_fields.add(extra_field_str)
        if conversion_func:
            self._extra_fields_conversion_funcs[field_name] = conversion_func
        return self

    """QUERY BUILDING METHODS"""

    def _filter_str(self) -> str:
        if not self._filter:
            return ""
        return f"FILTER {self._filter}"

    def _order_by_str(self) -> str:
        if not self._order_by:
            return ""
        return f"ORDER BY {self._order_by}"

    def _limit_str(self) -> str:
        if not self._limit or self._limit == 0:
            return ""
        return f"LIMIT {self._limit}"

    def _offset_str(self) -> str:
        if not self._offset or self._offset == 0:
            return ""
        return f"OFFSET {self._offset}"

    def build_filters_str_and_vars(self, prefix: str) -> tuple[str, VARS]:
        """Only returning the vars for THIS obj"""
        s_lst = [
            self._filter_str(),
            self._order_by_str(),
            self._offset_str(),
            self._limit_str(),
        ]
        s = " ".join([s for s in s_lst if s])
        if prefix:
            # regex out the vars to include this prefix
            new_prefix = f"{prefix}{helpers.SEPARATOR}"
            new_s = s.replace("$", f"${new_prefix}")
            new_vars = {f"{new_prefix}{k}": v for k, v in self._query_variables.items()}
            return new_s, new_vars
        else:
            return s, self._query_variables

    def full_query_str_and_vars(
        self,
        include_select: bool,
        prefix: str,
        include_filters: bool = True,
        include_detached: bool = False,
        check_for_intersecting_variables: bool = False,
        model_name_override: str = None,
    ) -> tuple[str, VARS]:
        self.merge()
        model_name = model_name_override or self.model_name
        detached_str = f" DETACHED" if include_detached else ""
        select = f"SELECT{detached_str} {model_name} " if include_select else ""
        (
            nested_query_str,
            nested_vars,
        ) = self._nested_resolvers.build_query_str_and_vars(prefix=prefix)

        brackets_strs = [*self._fields_to_return, *self._extra_fields, nested_query_str]
        brackets_str = ", ".join(sorted([s for s in brackets_strs if s]))
        s = f"{select}{{ {brackets_str} }}"

        if include_filters:
            filters_str, query_vars = self.build_filters_str_and_vars(prefix=prefix)
            if filters_str:
                s += f" {filters_str}"
        else:
            query_vars = {}

        if check_for_intersecting_variables:
            # this is unlikely to happen because of the separator and prefix but just for sanity you can do this
            # if you do not have "__" in your variables this *is* impossible
            if inters := (query_vars.keys() & nested_vars.keys()):
                for var_name in inters:
                    if query_vars[var_name] != nested_vars[var_name]:
                        raise errors.ResolverException(
                            f"Variable {var_name} was given multiple times with different values: "
                            f"{query_vars[var_name]} != {nested_vars[var_name]}"
                        )

        return s, {**query_vars, **nested_vars}

    """MERGING LOGIC"""

    def build_hydrated_filters_str(self) -> str:
        filters_str, _ = self.build_filters_str_and_vars(prefix="")
        return helpers.replace_str_with_vars(
            s=filters_str, variables=self._query_variables
        )

    def is_subset_of(self, other: "Resolver", should_debug: bool = False) -> bool:  # type: ignore
        if self is other:
            return True
        if self._fields_to_return:
            self_additional_fields_to_return = (
                self._fields_to_return - other._fields_to_return
            )
            if self_additional_fields_to_return:
                logger.debug(f"{self_additional_fields_to_return=}")
                return False

        if self._extra_fields:
            self_additional_extra_fields = self._extra_fields - other._extra_fields
            if self_additional_extra_fields:
                logger.debug(f"{self_additional_extra_fields=}")
                return False
            self_additional_conversion_funcs = (
                self._extra_fields_conversion_funcs.keys()
                - other._extra_fields_conversion_funcs
            )
            if self_additional_conversion_funcs:
                logger.debug(f"{self_additional_conversion_funcs=}")
                return False

        # compare filter strs then variables then nested
        for key, val in self._query_variables.items():
            if key not in other._query_variables:
                logger.debug(f"{key} not in other._query_variables")
                return False
            if other._query_variables[key] != val:
                logger.debug(f"{other._query_variables[key]=} != {val=}")
                return False

        # PROs of this... it should be very safe since you are comparing the actual VARS
        # cons, it could be overly restrictive. If one is called $start_time vs $startTime it will break...fine tho
        self_filters_str, _ = self.build_filters_str_and_vars(prefix="")
        other_filters_str, _ = other.build_filters_str_and_vars(prefix="")
        if self_filters_str != other_filters_str:
            if should_debug:
                logger.debug(
                    f"{self.__class__.__name__}: {self_filters_str=} != {other_filters_str}"
                )
            return False

        if not self._nested_resolvers.is_subset_of(other._nested_resolvers):
            logger.debug(
                f"self nested_resolvers are not subset of other nested_resolvers"
            )
            return False

        return True

    """QUERY METHODS"""

    async def query(
        self, client: edgedb.AsyncIOClient | None = None
    ) -> T.List[NodeType]:
        query_str, variables = self.full_query_str_and_vars(
            include_select=True, prefix=""
        )
        with span.span(
            op=f"edgedb.query.{self.model_name}", description=query_str[:200]
        ):
            raw_response = await execute.query(
                client=client or self._node_config.client,
                query_str=query_str,
                variables=variables,
                only_one=False,
            )
        if not isinstance(raw_response, list):
            raise errors.ResolverException(
                f"Expected a list from query, got {raw_response}."
            )
        return self.parse_obj_with_cache_list(raw_response)

    async def query_first(
        self, client: edgedb.AsyncIOClient | None = None
    ) -> NodeType | None:
        if self._limit is not None and self._limit > 1:
            raise errors.ResolverException(
                f"Limit is set to {self._limit} so you cannot query_first."
            )
        self._limit = 1
        model_lst = await self.query(client=client)
        if not model_lst:
            return None
        return model_lst[0]

    async def _get(
        self,
        field_name: str,
        value: T.Any,
        *,
        client: edgedb.AsyncIOClient | None = None,
    ) -> NodeType | None:
        self.validate_field_name_value_filters(
            operation_name="get", field_name=field_name, value=value
        )
        query_str, variables = self.full_query_str_and_vars(
            include_select=True, prefix=""
        )
        custom_filter_str = f"FILTER {self.filter_str_from_field_name(field_name)}"
        query_str += f" {custom_filter_str}"
        with span.span(op=f"edgedb.get.{self.model_name}", description=query_str[:200]):
            raw_response = await execute.query(
                client=client or self._node_config.client,
                query_str=query_str,
                variables={**variables, field_name: value},
                only_one=True,
            )

        if not raw_response:
            return None
        return self.parse_obj_with_cache(raw_response)

    async def _gerror(
        self,
        field_name: str,
        value: T.Any,
        *,
        client: edgedb.AsyncIOClient | None = None,
    ) -> NodeType:
        model = await self._get(field_name=field_name, value=value, client=client)
        if not model:
            raise errors.ResolverException(
                f"No {self.model_name} in db with fields {field_name} = {value}."
            )
        return model

    """MUTATION METHODS"""

    @staticmethod
    def validate_upsert_fields(
        upsert_given_conflict_on: str = None,
        return_model_for_conflict_on: str = None,
        custom_conflict_on_str: str = None,
    ) -> None:
        if custom_conflict_on_str:
            if return_model_for_conflict_on or upsert_given_conflict_on:
                raise errors.ResolverException(
                    "You cannot give a custom conflict string with other conflict inputs."
                )
        if return_model_for_conflict_on and upsert_given_conflict_on:
            raise errors.ResolverException(
                "You cannot both suppress a conflict and upsert given a conflict."
            )

    def build_conflict_str_and_vars(
        self,
        insert: InsertType,
        *,
        mutate_on_update: bool,
        custom_conflict_on_str: str | None,
        upsert_given_conflict_on: str | None,
        return_model_for_conflict_on: str | None,
    ) -> tuple[str, VARS]:
        self.validate_upsert_fields(
            upsert_given_conflict_on=upsert_given_conflict_on,
            return_model_for_conflict_on=return_model_for_conflict_on,
            custom_conflict_on_str=custom_conflict_on_str,
        )

        conflict_str = ""
        conflict_variables: VARS = {}

        if custom_conflict_on_str:
            conflict_str = custom_conflict_on_str
        elif upsert_given_conflict_on:
            patch = self.patch_from_insert(insert)
            conflict_s, conflict_variables = utils.model_to_set_str_vars(
                model=patch,
                conversion_map=self._node_config.patch_edgedb_conversion_map,
                additional_link_str=self.build_mutate_on_update_str(
                    patch=patch, mutate_on_update=mutate_on_update
                ),
            )
            conflict_str = f"UNLESS CONFLICT ON .{upsert_given_conflict_on} else (UPDATE {self.model_name} SET {conflict_s})"
        elif return_model_for_conflict_on:
            conflict_str = f"UNLESS CONFLICT ON .{return_model_for_conflict_on} ELSE (SELECT {self.model_name})"

        return conflict_str, conflict_variables

    async def insert_one(
        self,
        insert: InsertType,
        *,
        client: edgedb.AsyncIOClient | None = None,
        upsert_given_conflict_on: str = None,
        custom_conflict_on_str: str = None,
        return_model_for_conflict_on: str = None,
        mutate_on_update: bool = True,
    ) -> NodeType:
        if existing_filter_str := self.has_filters():
            raise errors.ResolverException(
                f"This resolver already has filters: {existing_filter_str}. "
                f"If you wish to INSERT an object, use a resolver that does not have root filters."
            )
        insert_s, insert_variables = utils.model_to_set_str_vars(
            model=insert, conversion_map=self._node_config.insert_edgedb_conversion_map
        )
        insert_s = f"INSERT {self.model_name} {insert_s}"

        conflict_str, conflict_variables = self.build_conflict_str_and_vars(
            insert=insert,
            upsert_given_conflict_on=upsert_given_conflict_on,
            custom_conflict_on_str=custom_conflict_on_str,
            return_model_for_conflict_on=return_model_for_conflict_on,
            mutate_on_update=mutate_on_update,
        )
        if conflict_str:
            insert_s += f" {conflict_str}"

        # do not need the prefix since any var HAS to be nested, so will already have prefixes
        select_s, select_variables = self.full_query_str_and_vars(
            prefix="", model_name_override="model", include_select=True
        )
        final_insert_s = f"WITH model := ({insert_s}) {select_s}"
        with span.span(op=f"edgedb.add.{self.model_name}"):
            raw_response = await execute.query(
                client=client or self._node_config.client,
                query_str=final_insert_s,
                variables={
                    **select_variables,
                    **insert_variables,
                    **conflict_variables,
                },
                only_one=True,
            )
        raw_response = T.cast(RAW_RESP_ONE, raw_response)
        return self.parse_obj_with_cache(raw_response)

    async def insert_many(
        self, inserts: list[InsertType], *, client: edgedb.AsyncIOClient | None = None
    ) -> list[NodeType]:
        if not inserts:
            return []
        conversion_map = self._node_config.insert_edgedb_conversion_map
        first_insert_s, _ = utils.model_to_set_str_vars(
            model=inserts[0], conversion_map=conversion_map, json_get_item="item"
        )
        insert_vars_list: list[VARS] = []
        for insert in inserts:
            # confirm that the INSERT STRS of all of these are the same
            insert_s, insert_vars = utils.model_to_set_str_vars(
                model=insert, conversion_map=conversion_map, json_get_item="item"
            )
            if insert_s != first_insert_s:
                raise errors.ResolverException(
                    f"Not all inserts have the same form: {insert_s} != {first_insert_s}."
                )
            insert_vars_list.append(insert_vars)

        insert_s = f"INSERT {self.model_name} {first_insert_s}"
        select_s, select_variables = self.full_query_str_and_vars(
            prefix="", model_name_override="model", include_select=False
        )

        # for insert_s, replace $x with json_get(item, "x")
        insert_s = re.sub(
            pattern=r"\$(\w+)", repl='json_get(item, "' + r"\1" + '")', string=insert_s
        )

        final_insert_str = f"""
        with
            raw_data := <json>$__data,
        for item in json_array_unpack(raw_data) union ({insert_s}) {select_s}
                """
        variables = {
            **select_variables,
            "__data": json.dumps(encoders.jsonable_encoder(insert_vars_list)),
        }
        debug(variables)
        with span.span(op=f"edgedb.add_many.{self.model_name}"):
            raw_response = await execute.query(
                client=client or self._node_config.client,
                query_str=final_insert_str,
                variables=variables,
                only_one=False,
            )
        raw_response = T.cast(RAW_RESP_MANY, raw_response)
        return self.parse_obj_with_cache_list(raw_response)

    def build_mutate_on_update_str(
        self, patch: PatchType, mutate_on_update: bool | None
    ) -> str | None:
        if mutate_on_update is False:
            return None
        mutate_on_update_d = self._node_config.mutate_on_update
        if not mutate_on_update_d:
            return None
        mutate_on_update_strs: list[str] = []
        for field_name, expression in mutate_on_update_d.items():
            if field_name not in patch.__fields_set__:
                mutate_on_update_strs.append(f"{field_name} := {expression}")
        mutate_on_update_str = ", ".join(mutate_on_update_strs)
        return mutate_on_update_str

    async def _update(
        self,
        patch: PatchType,
        *,
        field_name: str = None,
        value: T.Any = None,
        only_one: bool,
        mutate_on_update: bool,
        client: edgedb.AsyncIOClient | None,
    ) -> RAW_RESPONSE:
        update_s, update_variables = utils.model_to_set_str_vars(
            model=patch,
            conversion_map=self._node_config.patch_edgedb_conversion_map,
            additional_link_str=self.build_mutate_on_update_str(
                patch=patch, mutate_on_update=mutate_on_update
            ),
        )
        filters_s, filters_vars = self.build_filters_str_and_vars(prefix="")

        if only_one:
            if field_name is None:
                raise errors.ResolverException(
                    "Cannot be only_one and have no field_name."
                )
            custom_filter_str = f"FILTER {self.filter_str_from_field_name(field_name)}"
            filters_s += f" {custom_filter_str}"
            filters_vars[field_name] = value

        update_s = f"UPDATE {self.model_name} {filters_s} SET {update_s}"
        select_s, select_variables = self.full_query_str_and_vars(
            prefix="",
            model_name_override="model",
            include_select=True,
            include_filters=False,
        )
        final_update_s = f"WITH model := ({update_s}) {select_s}"
        with span.span(op=f"edgedb.update.{self.model_name}"):
            raw_response = await execute.query(
                client=client or self._node_config.client,
                query_str=final_update_s,
                variables={**select_variables, **filters_vars, **update_variables},
                only_one=only_one,
            )
        raw_response = T.cast(RAW_RESPONSE, raw_response)
        return raw_response

    def validate_field_name_value_filters(
        self, operation_name: str, field_name: str, value: T.Any
    ) -> None:
        if existing_filter_str := self.has_filters():
            raise errors.ResolverException(
                f"`{operation_name}` requires a resolver with *no* root filters but this resolver has root filters: "
                f"{existing_filter_str}. Instead, pass in the exclusive field + value "
                f"or use a resolver without root filters."
            )
        if value is None:
            raise errors.ResolverException("Value must not be None.")
        if field_name not in self._node_config.exclusive_fields:
            raise errors.ResolverException(f"Field '{field_name}' is not exclusive.")

    async def _update_one(
        self,
        patch: PatchType,
        *,
        field_name: str,
        value: T.Any,
        client: edgedb.AsyncIOClient | None = None,
        mutate_on_update: bool = True,
    ) -> NodeType:
        if not patch.__fields_set__:
            raise errors.ResolverException("Patch is empty.")
        self.validate_field_name_value_filters(
            operation_name="update_one", field_name=field_name, value=value
        )
        raw_response = await self._update(
            patch=patch,
            field_name=field_name,
            value=value,
            only_one=True,
            client=client,
            mutate_on_update=mutate_on_update,
        )
        raw_response = T.cast(RAW_RESP_ONE, raw_response)
        if not raw_response:
            raise errors.ResolverException("No object to update.")
        return self.parse_obj_with_cache(raw_response)

    async def update_many(
        self,
        patch: PatchType,
        *,
        update_all: bool = False,
        client: edgedb.AsyncIOClient | None = None,
        mutate_on_update: bool = True,
    ) -> list[NodeType]:
        if not patch.__fields_set__:
            raise errors.ResolverException("Patch is empty.")
        if not update_all:
            if not self.has_filters():
                raise errors.ResolverException(
                    "You did not give filters which means this will update *all* models. "
                    "If this is your intention, pass update_all=True."
                )
        raw_response = await self._update(
            patch=patch,
            only_one=False,
            client=client,
            mutate_on_update=mutate_on_update,
        )
        raw_response = T.cast(RAW_RESP_MANY, raw_response)
        return self.parse_obj_with_cache_list(raw_response)

    async def _delete(
        self,
        *,
        field_name: str = None,
        value: T.Any = None,
        only_one: bool,
        client: edgedb.AsyncIOClient | None,
    ) -> RAW_RESPONSE:
        filters_s, filters_vars = self.build_filters_str_and_vars(prefix="")

        if only_one:
            if field_name is None:
                raise errors.ResolverException(
                    "Cannot be only_one and have no field_name."
                )
            custom_filter_str = f"FILTER {self.filter_str_from_field_name(field_name)}"
            filters_s += f" {custom_filter_str}"
            filters_vars[field_name] = value

        delete_s = f"DELETE {self.model_name} {filters_s}"
        select_s, select_variables = self.full_query_str_and_vars(
            prefix="",
            model_name_override="model",
            include_select=True,
            include_filters=False,
        )
        final_delete_s = f"WITH model := ({delete_s}) {select_s}"
        with span.span(op=f"edgedb.delete.{self.model_name}"):
            raw_response = await execute.query(
                client=client or self._node_config.client,
                query_str=final_delete_s,
                variables={**select_variables, **filters_vars},
                only_one=only_one,
            )
        raw_response = T.cast(RAW_RESPONSE, raw_response)
        return raw_response

    async def _delete_one(
        self,
        *,
        field_name: str,
        value: T.Any,
        client: edgedb.AsyncIOClient | None = None,
    ) -> NodeType:
        self.validate_field_name_value_filters(
            operation_name="delete one", field_name=field_name, value=value
        )
        raw_response = await self._delete(
            field_name=field_name, value=value, only_one=True, client=client
        )
        raw_response = T.cast(RAW_RESP_ONE, raw_response)
        if not raw_response:
            raise errors.ResolverException("No object to delete.")
        return self.parse_obj_with_cache(raw_response)

    async def delete_many(
        self,
        *,
        delete_all: bool = False,
        client: edgedb.AsyncIOClient | None = None,
    ) -> list[NodeType]:
        if not delete_all:
            if not self.has_filters():
                raise errors.ResolverException(
                    "You did not give filters which means this will delete *all* models. "
                    "If this is your intention, pass delete_all=True."
                )
        raw_response = await self._delete(only_one=False, client=client)
        raw_response = T.cast(RAW_RESP_MANY, raw_response)
        return self.parse_obj_with_cache_list(raw_response)

    """HELPERS"""

    def has_filters(self) -> T.Optional[str]:
        for field in FILTER_FIELDS:
            if (val := getattr(self, field)) is not None:
                return val
        return None

    def patch_from_insert(self, insert: InsertType) -> PatchType:
        patch = self._patch_cls()
        for field in insert.__fields_set__:
            if field in self._patch_cls.__fields__:
                setattr(patch, field, getattr(insert, field))
        return patch

    """PARSING"""

    def _parse_obj_with_cache(self, d: RAW_RESP_ONE) -> NodeType:
        # TODO counts will fail, catch counts early
        node = self._node_cls(**d)
        # TODO speed test
        fields_set = {re.sub(r"_$", "", s) for s in node.__fields_set__}
        other_fields = d.keys() - fields_set
        for field_name in sorted(other_fields):
            resolver = self._nested_resolvers.resolver_from_field_name(field_name)
            if not resolver:
                # must be an extra field
                node.computed[field_name] = d[field_name]
            else:
                child = d[field_name]
                if child:
                    if isinstance(child, list):
                        val = [resolver._parse_obj_with_cache(d) for d in child]
                    elif isinstance(child, int):
                        # counts
                        val = child
                    else:
                        val = resolver._parse_obj_with_cache(child)
                else:
                    val = child
                edge_name = field_name.split(helpers.SEPARATOR)[0]
                node._cache.add(edge=edge_name, resolver=resolver, val=val)
        node._used_resolver = self
        return node

    def parse_obj_with_cache(self, d: RAW_RESP_ONE) -> NodeType:
        with span.span(op=f"parse.{self.model_name}"):
            return self._parse_obj_with_cache(d)

    def parse_obj_with_cache_list(self, lst: RAW_RESP_MANY) -> list[NodeType]:
        with span.span(op=f"parse_list.{self.model_name}", description=f"{len(lst)}"):
            return [self._parse_obj_with_cache(d) for d in lst]

    """merge"""

    def merge(self) -> None:
        self._nested_resolvers = merge_nested_resolver(self._nested_resolvers)



from .main import DBConfig, DBVendor, NodeConfig, generate, PropertyConfig
from edge_orm.node import PropertyCardinality

__all__ = [
    "DBConfig",
    "DBVendor",
    "generate",
    "NodeConfig",
    "PropertyConfig",
    "PropertyCardinality",
]



# TODO copied from legacy so it is messy
import typing as T
from .nested_resolvers import NestedResolvers
from edge_orm import logger

if T.TYPE_CHECKING:
    from .model import Resolver

ResolverType = T.TypeVar("ResolverType", bound="Resolver")  # type: ignore


class MergeException(Exception):
    pass


def merge_fields(
    a: ResolverType,
    b: ResolverType,
    merged_resolver: ResolverType,
    key: str,
    separator: str = ", ",
) -> None:
    a_val = getattr(a, key, None)
    b_val = getattr(b, key, None)

    if a_val == b_val:
        setattr(merged_resolver, key, a_val)
    elif a_val is None or b_val is None:
        # one is None, so just use the other one's
        setattr(merged_resolver, key, a_val or b_val)
    else:
        # they both are not none and also not the same, hope there is no conflict!
        setattr(merged_resolver, key, f"{a_val}{separator}{b_val}")


def merge_resolvers(
    a: ResolverType, b: ResolverType, should_debug: bool = False
) -> T.Optional[ResolverType]:
    # first merge, then see if a and b are subsets. If they are, return! else, None, they can't be merged
    if a.__class__ != b.__class__:
        raise MergeException(
            f"Resolvers are not the same type, {a.__class__=}, {b.__class__=}"
        )
    merged_resolver = a.__class__()
    merged_resolver._merged = True

    # for these, filters must be the same anyway
    merged_resolver._filter = a._filter
    merged_resolver._order_by = a._order_by
    merged_resolver._limit = a._limit
    merged_resolver._offset = a._offset

    merge_fields(a, b, merged_resolver, key="_extra_fields")
    # merge_fields(a, b, merged_resolver, key="_modules")

    # can ignore update operation? -> yeah because these aren't used for updating
    merged_resolver._query_variables = {**a._query_variables, **b._query_variables}

    if a._fields_to_return is None and b._fields_to_return is None:
        merged_resolver._fields_to_return = None
    else:
        merged_resolver._fields_to_return = {
            *(a._fields_to_return or set()),
            *(b._fields_to_return or set()),
        }
    merged_resolver._fields_to_return = {*a._fields_to_return, *b._fields_to_return}

    merged_resolver._nested_resolvers = merge_nested_resolvers(
        a._nested_resolvers, b._nested_resolvers
    )

    # okay, now if subsets, return ! else None
    # or separate this??
    if a.is_subset_of(merged_resolver) and b.is_subset_of(merged_resolver):
        return merged_resolver
    if should_debug:
        logger.warning("NOT SUBSETS!")
        logger.warning(
            f"{a.is_subset_of(merged_resolver)=}, {b.is_subset_of(merged_resolver)=}"
        )
    return None


def merge_nested_resolvers(
    a: NestedResolvers, b: NestedResolvers
) -> T.Optional[NestedResolvers]:
    merged_nested_resolvers = NestedResolvers()
    all_edges = {*a.d.keys(), *b.d.keys()}

    for edge in all_edges:
        merged_resolvers: list[Resolver] = []  # type: ignore
        has_merged: list[Resolver] = []  # type: ignore

        a_resolvers: list[Resolver] = a.get(edge)  # type: ignore
        b_resolvers: list[Resolver] = b.get(edge)  # type: ignore
        for a_resolver in a_resolvers:
            if a_resolver in has_merged:
                continue
            for b_resolver in b_resolvers:
                if b_resolver in has_merged:
                    continue
                if merged_resolver := merge_resolvers(a_resolver, b_resolver):
                    merged_resolvers.append(merged_resolver)
                    # now mark these two resolvers as merged
                    has_merged.append(a_resolver)
                    has_merged.append(b_resolver)
        # do a second pass, all resolvers not merged add to resolver list
        for r in [*a_resolvers, *b_resolvers]:
            if r not in has_merged:
                merged_resolvers.append(r)

        # for r in merged_resolvers:
        #     merged_nested_resolvers.add(edge, resolver=r)
        if edge not in merged_nested_resolvers.d:
            merged_nested_resolvers.d[edge] = []
        merged_nested_resolvers.d[edge] = merged_resolvers

    return merged_nested_resolvers


def merge_resolvers_lst(resolvers: list[ResolverType]) -> list[ResolverType]:
    if not resolvers:
        return []
    q = resolvers.copy()
    merged: list[ResolverType] = []
    while q:
        new_q: list[ResolverType] = []
        main_r = q.pop()
        for other_r in q:
            if merged_r := merge_resolvers(main_r, other_r):
                main_r = merged_r
            else:
                new_q.append(other_r)
        merged.append(main_r)
        q = new_q
    return merged


def merge_nested_resolver(nested_resolvers: NestedResolvers) -> NestedResolvers:
    """
    merged_nested_resolvers = NestedResolvers()
    for edge, resolvers in nested_resolvers.d.items():
        merged_nested_resolvers.d[edge] = merge_resolvers_lst(resolvers)
    return merged_nested_resolvers
    """
    merged_nested_resolvers = NestedResolvers()
    for edge, resolvers in nested_resolvers.d.items():
        for r in resolvers:
            r.merge()
            merged_nested_resolvers.add(edge=edge, resolver=r, merge=True)
    return merged_nested_resolvers



import time
import typing as T
from contextlib import contextmanager
from edge_orm import logger

try:
    from sentry_sdk import start_span

    USE_SENTRY = True
except ModuleNotFoundError:
    logger.debug("Sentry not found, will not span.")
    USE_SENTRY = False


@contextmanager
def timer(*, op: str, description: str = None, use: bool) -> T.Iterator[T.Any]:
    if use:
        start = time.time()
        yield
        took_ms = round((time.time() - start) * 1000, 2)
        logger.debug(f"{op=}, {description} took {took_ms} ms")
    else:
        yield


@contextmanager
def span(
    op: str,
    description: str = None,
    use: bool = True,
    log_time: bool = False,
    **kwargs: T.Any,
) -> T.Iterator[T.Any]:
    if USE_SENTRY is False or use is False:
        yield
    else:
        with timer(op=op, description=description, use=log_time):
            with start_span(op=op, description=description, **kwargs):
                yield



import typing as T
import time
import orjson
import edgedb
from edge_orm import logger
from edge_orm.span import span


class ExecuteException(Exception):
    pass


class ExecuteConstraintViolationException(ExecuteException):
    pass


MUTATION_ACTIONS = {"insert ", "update ", "delete "}


def operation_from_query_str(query_str: str) -> str:
    s = query_str.lower()
    for action in MUTATION_ACTIONS:
        if action in s:
            return "mutation"
    return "query"


async def query(
    *,
    client: edgedb.AsyncIOClient,
    query_str: str,
    variables: dict[str, T.Any] | None = None,
    only_one: bool,
) -> T.Any | None:
    if variables is None:
        variables = {}
    # TODO usually would simplify vars here but should do this in earlier step
    query_func = client.query_json if not only_one else client.query_single_json
    start = time.time()
    try:
        with span(
            op=f"edge-orm.{operation_from_query_str(query_str)}",
            description=query_str[:200],
        ):
            json_str = await query_func(query=query_str, **variables)
        with span(op=f"orjson.loads", description=f"{len(json_str)=}"):
            response_dict = orjson.loads(json_str)
    except edgedb.errors.ConstraintViolationError as e:
        logger.error(f"{e=}")
        if "is prohibited by link target policy" in str(e):
            raise e
        if "violates exclusivity constraint" in str(e):
            field_name = str(e).split(" ")[0].replace("_", " ")
            raise ExecuteConstraintViolationException(
                f"That {field_name} already exists in our system."
            )
        raise e
    except Exception as e:
        logger.error(
            f"EdgeDB Query Exception: {e}, query_str and variables: {query_str=}, {variables=}"
        )
        raise e
    took_ms = round((time.time() - start) * 1_000, 2)
    logger.debug(query_str)
    logger.debug(f"took {took_ms} ms")
    return response_dict



from .errors import NodeException
from .models import (
    Node,
    Insert,
    Patch,
    CONVERSION_MAP,
    EdgeConfigBase,
    PropertyCardinality,
)

__all__ = [
    "NodeException",
    "Node",
    "Insert",
    "Patch",
    "CONVERSION_MAP",
    "EdgeConfigBase",
    "PropertyCardinality",
]



import typing as T
from pydantic import BaseModel
from edge_orm import helpers
from edge_orm.types_generator.main import COUNT_POSTFIX

if T.TYPE_CHECKING:
    from .model import Resolver, VARS

ResolverType = T.TypeVar("ResolverType", bound="Resolver")


class NestedResolvers(BaseModel):
    d: dict[str, list[T.Any]] = {}

    def get(self, edge: str) -> list[ResolverType]:
        return self.d.get(edge, [])

    def has(self, edge: str) -> bool:
        return edge in self.d

    def add(
        self,
        edge: str,
        resolver: ResolverType,
        *,
        merge: bool = False,
        make_first: bool = False,
    ) -> None:
        from .merging import merge_resolvers

        if not self.has(edge):
            self.d[edge] = []

        if merge:
            existing_resolvers: list[ResolverType] = self.get(edge)
            new_resolvers: list[ResolverType] = []
            has_merged = False
            for existing_r in existing_resolvers:
                if has_merged is False and (
                    merged_r := merge_resolvers(resolver, existing_r)
                ):
                    new_resolvers.append(merged_r)
                    has_merged = True
                else:
                    new_resolvers.append(existing_r)
            self.d[edge] = new_resolvers
            if has_merged is True:
                return

        if make_first:
            self.d[edge].insert(0, resolver)
        else:
            self.d[edge].append(resolver)

    def has_subset(self, edge: str, resolver: ResolverType) -> bool:
        for r in self.get(edge):  # type: ignore
            if resolver.is_subset_of(r):
                return True
        return False

    def is_subset_of(self, other: "NestedResolvers") -> bool:
        for edge, resolvers in self.d.items():
            for r in resolvers:
                if not other.has_subset(edge=edge, resolver=r):
                    return False
        return True

    def edge_to_query_str_and_vars(self, edge: str, prefix: str) -> tuple[str, "VARS"]:
        resolvers: list["Resolver"] = self.get(edge)
        resolvers_str = []
        vars_lst = []
        for i, r in enumerate(resolvers):
            if i == 0:
                new_prefix = f"{prefix}{helpers.SEPARATOR}{edge}" if prefix else edge
                if r.is_count and COUNT_POSTFIX in edge:
                    # avoid not copying resolver and having it break. make SURE it is a count
                    filters_str, variables = r.build_filters_str_and_vars(
                        prefix=new_prefix
                    )
                    resolver_s = f"{edge} := count((SELECT .{edge.split(COUNT_POSTFIX)[0]} {filters_str}))"
                else:
                    filters_str, variables = r.full_query_str_and_vars(
                        include_select=False, prefix=new_prefix
                    )
                    resolver_s = f"{edge}: {filters_str}"
            else:
                key_name = f"{edge}{helpers.SEPARATOR}{i}"
                new_prefix = (
                    f"{prefix}{helpers.SEPARATOR}{key_name}" if prefix else key_name
                )
                if r.is_count and COUNT_POSTFIX in edge:
                    filters_str, variables = r.build_filters_str_and_vars(
                        prefix=new_prefix
                    )
                    resolver_s = f"{key_name} := count((SELECT .{edge.split(COUNT_POSTFIX)[0]} {filters_str}))"
                else:
                    filters_str, variables = r.full_query_str_and_vars(
                        include_select=False, prefix=new_prefix
                    )
                    resolver_s = f"{key_name} := (SELECT .{edge} {filters_str})"
            resolvers_str.append(resolver_s)
            vars_lst.append(variables)
        flattened_d = {k: v for d in vars_lst for k, v in d.items()}
        return ", ".join(resolvers_str), flattened_d

    def build_query_str_and_vars(self, prefix: str) -> tuple[str, "VARS"]:
        edge_strs: list[str] = []
        vars_lst: list["VARS"] = []
        keys = self.d.keys()
        for i, edge in enumerate(keys):
            s, v = self.edge_to_query_str_and_vars(edge=edge, prefix=prefix)
            edge_strs.append(s)
            vars_lst.append(v)

        s = ", ".join(edge_strs)
        flattened_d = {k: v for d in vars_lst for k, v in d.items()}
        return s, flattened_d

    def resolver_from_field_name(self, field_name: str) -> ResolverType | None:
        possible_edge = field_name.split(helpers.SEPARATOR)[0]
        resolvers: list[ResolverType] = self.get(possible_edge)
        if not resolvers:
            return None
        if helpers.SEPARATOR not in field_name:
            return resolvers[0]
        try:
            index = int(field_name.split(helpers.SEPARATOR)[1])
            return resolvers[index]
        except (IndexError, ValueError):
            return None

    """MERGE"""

    def merge(self) -> "NestedResolvers":
        merged_nested_resolvers = NestedResolvers()
        for edge, resolvers in self.d.items():
            for r in resolvers:
                r.merge()
                merged_nested_resolvers.add(edge=edge, resolver=r, merge=True)
        return merged_nested_resolvers



from typing import Any, Optional, Type


class UnsetType:
    """Taken from Strawberry: https://strawberry.rocks/"""

    __instance: Optional["UnsetType"] = None

    def __new__(cls: Type["UnsetType"]) -> "UnsetType":
        if cls.__instance is None:
            ret = super().__new__(cls)
            cls.__instance = ret
            return ret
        else:
            return cls.__instance

    def __str__(self) -> str:
        return ""

    def __repr__(self) -> str:
        return "UNSET"

    def __bool__(self) -> bool:
        return False


UNSET: Any = UnsetType()

__all__ = ["UNSET", "UnsetType"]

# UNSET = "*_~_**"
# __all__ = ["UNSET"]



import re
import typing as T
from enum import Enum
import os
from pathlib import Path
from black import format_str, FileMode
import edgedb
from pydantic import BaseModel, parse_raw_as, Field
from .introspection import (
    introspect_objects,
    introspect_scalars,
    ObjectType,
    Link,
    Property,
)
from edge_orm.node.models import CONVERSION_MAP, PropertyCardinality, FieldInfo

ENV_VAR_PATTERN = r"[A-Z_]+"
COUNT_POSTFIX = "_Count"


class DBVendor(str, Enum):
    edgedb = "edgedb"


class PropertyConfig(BaseModel):
    module_path: str
    module_name: str
    validate_as_basemodel: bool = True
    cardinality: PropertyCardinality = PropertyCardinality.ONE


class NodeConfig(BaseModel):
    module_path: str | None = None
    appendix_properties: T.List[str] = []
    ignore_properties: T.List[str] = []
    basemodel_properties: T.Dict[str, PropertyConfig] = {}
    custom_annotations: T.Dict[str, str] = {}
    mutate_on_update: T.Dict[str, str] = {}


class DBConfig(BaseModel):
    vendor: DBVendor
    dsn_: str = Field(..., alias="dsn")
    copy_config: str | None = None
    hydrate: bool = False
    nodes: T.Dict[str, NodeConfig] = dict()
    cache_only: bool = True
    resolver_mixin_path: str | None = None

    @property
    def is_plaintext_dsn(self) -> bool:
        return self.dsn_.startswith("edgedb://")

    @property
    def dsn(self) -> str:
        if self.is_plaintext_dsn:
            return self.dsn_
        return os.environ[self.dsn_]

    def dsn_str(self) -> str:
        if self.is_plaintext_dsn:
            return f'"{self.dsn_}"'
        return f'os.environ["{self.dsn_}"]'


PATH_TO_MODULE = "edge_orm"
DEFAULT_INDENT = "    "
CONFIG_NAME = "GraphORM"


class GeneratorException(Exception):
    pass


def indent_lines(s: str, indent: str = DEFAULT_INDENT) -> str:
    chunks = s.split("\n")
    return indent + f"\n{indent}".join(chunks)


def imports(enums_module: str, resolver_mixin_path: str | None) -> str:
    lines = [
        "from __future__ import annotations",
        "import os",
        "import typing as T",
        "from enum import Enum",
        "from datetime import datetime, date, timedelta",
        "from uuid import UUID",
        "from decimal import Decimal",
        "from edgedb import RelativeDuration, AsyncIOClient, create_async_client",
        "from pydantic import BaseModel, Field, PrivateAttr, validator",
        f"from {PATH_TO_MODULE}.node.models import Cardinality, FieldInfo, classproperty",
        f"from {PATH_TO_MODULE} import Node, Insert, Patch, EdgeConfigBase, Resolver, NodeException, ResolverException, UNSET, UnsetType, validators, errors, resolver_enums",
        "FilterConnector = resolver_enums.FilterConnector",
        f"from . import {enums_module} as enums",
    ]
    if resolver_mixin_path:
        lines.append(resolver_mixin_path)
    return "\n".join(lines)


def build_enum_imports() -> str:
    lines = ["from enum import Enum"]
    return "\n".join(lines)


async def build_enums(client: edgedb.AsyncIOClient, include_strawberry: bool) -> str:
    scalar_types = await introspect_scalars(client)
    enum_strs: T.List[str] = []
    enum_names: T.List[str] = []
    for scalar in scalar_types:
        if not scalar.enum_values:
            continue
        enum_value_strs: T.List[str] = [f'{e} = "{e}"' for e in scalar.enum_values]
        enum_value_str = "\n".join(enum_value_strs)
        s = f"class {scalar.node_name}(str, Enum):\n{indent_lines(enum_value_str)}"
        enum_strs.append(s)
        enum_names.append(scalar.node_name)
    if include_strawberry:
        straw_enums_str = "\n".join(
            [f"{enum_name} = strawberry.enum({enum_name})" for enum_name in enum_names]
        )
        straw_lines = [
            "import strawberry",
            f"class Strawberry:\n{indent_lines(straw_enums_str)}",
        ]
        enum_strs.extend(straw_lines)
    return "\n".join(enum_strs)


def build_node_link_function_str(link: Link) -> str:
    link_resolver_name = f"{link.target.model_name}Resolver"
    return f"""
async def {link.name}(
    self,
    resolver: {link_resolver_name} = None,
    cache_only: bool = CACHE_ONLY,
    client: AsyncIOClient | None = None,
) -> {link.type_str.replace("enums.", "")}:
    return await self.resolve(
        edge_name="{link.name}",
        edge_resolver=resolver or {link_resolver_name}(),
        cache_only=cache_only,
        client=client,
    )

async def {link.name}{COUNT_POSTFIX}(
    self,
    resolver: {link_resolver_name} = None,
    cache_only: bool = CACHE_ONLY,
    client: AsyncIOClient | None = None
) -> int:
    rez = resolver or {link_resolver_name}()
    rez.is_count = True
    return await self.resolve(
        edge_name="{link.name}{COUNT_POSTFIX}",
        edge_resolver=rez,
        cache_only=cache_only,
        client=client,
    )
    """


def build_resolver_link_function_str(node_resolver_name: str, link: Link) -> str:
    link_resolver_name = f"{link.target.model_name}Resolver"
    return f"""
def {link.name}(self, _: T.Optional[{link_resolver_name}] = None, /, make_first: bool = False) -> {node_resolver_name}:
    self._nested_resolvers.add("{link.name}", _ or {link_resolver_name}(), make_first=make_first)
    return self

def {link.name}{COUNT_POSTFIX}(self, _: T.Optional[{link_resolver_name}] = None, /, make_first: bool = False) -> {node_resolver_name}:
    rez = _ or {link_resolver_name}()
    rez.is_count = True
    self._nested_resolvers.add(
        "{link.name}{COUNT_POSTFIX}",
        rez,
        make_first=make_first
    )
    return self
    """


def build_exclusive_functions_str(
    node_name: str, exclusive_field_names: T.Set[str]
) -> str:
    exclusive_field_names_lst = sorted(list(exclusive_field_names))
    params_fields_str = ", ".join(
        [f"{f}: T.Optional[T.Any] = None" for f in exclusive_field_names_lst]
    )
    dict_fields_str = ", ".join([f'"{f}": {f}' for f in exclusive_field_names_lst])

    validation_str = f"""
kwargs = {{{dict_fields_str}}}
kwargs = {{k: v for k, v in kwargs.items() if v is not None}}
if len(kwargs) != 1:
    raise ResolverException(
        f"Must only give one argument, received {{kwargs}}."
    )
field_name, value = list(kwargs.items())[0]
""".strip()

    get_str = f"""
async def get(self, *, client: AsyncIOClient | None = None, {params_fields_str}) -> {node_name} | None:
{indent_lines(validation_str)}
    return await self._get(field_name=field_name, value=value, client=client)
    """
    gerror_str = f"""
async def gerror(self, *, client: AsyncIOClient | None = None, {params_fields_str}) -> {node_name}:
{indent_lines(validation_str)}
    return await self._gerror(field_name=field_name, value=value, client=client)
    """
    update_one_str = f"""
async def update_one(self, patch: {node_name}Patch, *, client: AsyncIOClient | None = None, {params_fields_str}) -> {node_name}:
{indent_lines(validation_str)}
    return await self._update_one(patch=patch, field_name=field_name, value=value, client=client)
    """
    delete_one_str = f"""
async def delete_one(self, *, client: AsyncIOClient | None = None, {params_fields_str}) -> {node_name}:
{indent_lines(validation_str)}
    return await self._delete_one(field_name=field_name, value=value, client=client)
    """
    return f"{get_str}\n{gerror_str}\n{update_one_str}\n{delete_one_str}\n"


def build_include_fields_function(
    node_resolver_name: str,
    appendix_properties: set[str],
    computed_properties: set[str],
) -> str:
    field_names = sorted({*appendix_properties, *computed_properties})
    names_params = ", ".join([f"{name}: bool = False" for name in field_names])
    logic_strs = "\n".join(
        [
            f'if {field_name} is True: fields_to_include.add("{field_name}")'
            for field_name in field_names
        ]
    )
    if not names_params:
        return ""
    s = f"""
def include(self, *, {names_params}) -> {node_resolver_name}:
    fields_to_include: set[str] = set()
{indent_lines(logic_strs)}
    return self.include_fields(*fields_to_include)
    """
    return s


def build_filter_functions_str(node_name: str, conversion_map: CONVERSION_MAP) -> str:
    field_names = sorted(conversion_map.keys())
    params_fields_str = ", ".join(
        [f"{f}: T.Optional[T.Any] = None" for f in field_names]
    )
    dict_fields_str = ", ".join([f'"{f}": {f}' for f in field_names])
    filter_by_str = f"""
def filter_by(self, filter_connector: FilterConnector = FilterConnector.AND, {params_fields_str}) -> {node_name}Resolver:
    return self._filter_by(connector=filter_connector, **{{{dict_fields_str}}})
    """
    # now for filter in
    params_fields_lst_str = ", ".join(
        [f"{f}: T.Optional[T.List[T.Any]] = None" for f in field_names]
    )
    filter_in_str = f"""
def filter_in(self, filter_connector: FilterConnector = FilterConnector.AND, {params_fields_lst_str}) -> {node_name}Resolver:
    return self._filter_in(connector=filter_connector, **{{{dict_fields_str}}})
    """
    return f"{filter_by_str}\n{filter_in_str}"


def add_quotes(lst: T.Iterable[str]) -> T.Iterable[str]:
    return [f'"{o}"' for o in lst]


def build_orm_config(
    model_name: str,
    updatable_fields: set[str],
    exclusive_fields: set[str],
    appendix_properties: set[str],
    computed_properties: set[str],
    basemodel_properties: T.Iterable[str],
    custom_annotations: T.Iterable[str],
    mutate_on_update: dict[str, str],
    node_edgedb_conversion_map: CONVERSION_MAP,
    insert_edgedb_conversion_map: CONVERSION_MAP,
    patch_edgedb_conversion_map: CONVERSION_MAP,
    insert_link_conversion_map: CONVERSION_MAP,
) -> str:
    return f"""
EdgeConfig: T.ClassVar[EdgeConfigBase] = EdgeConfigBase(
    model_name = "{model_name}",
    client = CLIENT,
    
    updatable_fields = {{{', '.join(add_quotes(sorted(list(updatable_fields))))}}},
    exclusive_fields = {{{', '.join(add_quotes(sorted(list(exclusive_fields))))}}},
    
    appendix_properties = {stringify_set(appendix_properties)},
    computed_properties = {stringify_set(computed_properties)},
    basemodel_properties = {stringify_set(set(basemodel_properties))},
    custom_annotations = {stringify_set(set(custom_annotations))},
    mutate_on_update = {stringify_dict(mutate_on_update)},
    
    node_edgedb_conversion_map = {stringify_basemodel_dict(node_edgedb_conversion_map)},
    insert_edgedb_conversion_map = {stringify_basemodel_dict(insert_edgedb_conversion_map)},
    patch_edgedb_conversion_map = {stringify_basemodel_dict(patch_edgedb_conversion_map)},
    
    insert_link_conversion_map = {stringify_basemodel_dict(insert_link_conversion_map)}
)
    """


def stringify_dict(
    d: T.Union[T.Dict[str, T.Any], str | bool], stringify_value: bool = True
) -> str:
    if type(d) is not dict:
        s = f"{d}"
        if type(d) is not bool:
            if stringify_value:
                s = f'"{s}"'
        return s
    inner = [
        f'"{k}":{stringify_dict(v, stringify_value=stringify_value)}'
        for k, v in d.items()
    ]
    return f"{{{','.join(inner)}}}"


def stringify_set(s: T.Set[str]) -> str:
    if not s:
        return "set()"
    s_sorted = sorted(list(s))
    strs: T.List[str] = [f'"{i}"' for i in s_sorted]
    return "{" + ",".join(strs) + "}"


def stringify_basemodel(model: BaseModel) -> str:
    "FieldInfo{cast: <str>} -> FieldInfo(cast='<str>')"
    str_lst: list[str] = []
    for k, v in model.dict().items():
        val_str = v
        if isinstance(v, Enum):
            val_str = f"{v.__class__.__name__}.{v.value}"
        elif isinstance(v, str):
            val_str = f'"{v}"'
        elif isinstance(v, BaseModel):
            val_str = stringify_basemodel(v)
        str_lst.append(f"{k}={val_str}")
    inner = ", ".join(str_lst)
    return f"{model.__class__.__name__}({inner})"


BaseModelType = T.TypeVar("BaseModelType", bound=BaseModel)


def stringify_basemodel_dict(d: dict[str, BaseModelType]) -> str:
    inner = [f'"{k}":{stringify_basemodel(v)}' for k, v in d.items()]
    return f"{{{','.join(inner)}}}"


def edgedb_conversion_type_from_prop(prop: Property) -> str:
    """
    s = prop.target.name
    pattern = r"default::\w+"
    s = re.sub(pattern, "std::str", s)
    return s
    """
    return prop.target.name


def build_validator_module_imports(db_config: DBConfig) -> str:
    if not db_config.nodes:
        return ""
    node_configs = db_config.nodes.values()
    from_str_import_strs: T.List[str] = []
    for node_config in node_configs:
        for prop_name, prop_config in node_config.basemodel_properties.items():
            from_str_import_strs.append(
                f"from {prop_config.module_path} import {prop_config.module_name}"
            )
    return "\n".join(sorted(list(set(from_str_import_strs))))


def key_from_field_name(f_name: str, node_config: NodeConfig) -> str:
    k = f_name
    if f_name in node_config.appendix_properties:
        k += "_"
    return f'"{k}"'


def build_from_str_validator_str(node_config: NodeConfig | None) -> str:
    if node_config is None:
        return ""
    field_name_strs: T.List[str] = []

    for field_name in node_config.custom_annotations.keys():
        field_name_strs.append(key_from_field_name(field_name, node_config=node_config))
    for field_name, props_config in node_config.basemodel_properties.items():
        if props_config.validate_as_basemodel is False:
            continue
        field_name_strs.append(key_from_field_name(field_name, node_config=node_config))
    if not field_name_strs:
        return ""
    return f"""
_from_str = validator({", ".join(field_name_strs)}, pre=True, allow_reuse=True)(validators.from_str)
    """


ListType = T.TypeVar("ListType")


def remove_falsies(lst: list[ListType]) -> list[ListType]:
    return [i for i in lst if i]


def build_node_and_resolver(
    object_type: ObjectType,
    node_config: T.Optional[NodeConfig],
    edge_resolver_map_strs: T.List[str],
    hydrate: bool,
    dehydrate: bool,
    allow_inserting_id: bool = True,
    resolver_mixin_model: str | None = None,
) -> str:
    # need to sort props and links by required, exclusive, no default, rest
    object_type.properties.sort(
        key=lambda x: f"{not x.is_computed}-{x.required}-{x.is_exclusive}-{x.default}",
        reverse=True,
    )
    object_type.links.sort(
        key=lambda x: f"{not x.is_computed}-{x.required}-{x.is_exclusive}-{x.default}",
        reverse=True,
    )
    # remove ignored properties
    if node_config:
        object_type.properties = [
            p
            for p in object_type.properties
            if p.name not in node_config.ignore_properties
        ]

    # start with the properties
    node_resolver_name = f"{object_type.node_name}Resolver"
    property_strs: T.List[str] = []
    insert_property_strs: T.List[str] = []
    patch_property_strs: T.List[str] = []
    updatable_fields: T.Set[str] = set()
    exclusive_fields: T.Set[str] = set()

    node_edgedb_conversion_map: CONVERSION_MAP = {}
    insert_edgedb_conversion_map: CONVERSION_MAP = {}
    patch_edgedb_conversion_map: CONVERSION_MAP = {}

    computed_properties: T.Set[str] = set()
    computed_property_getter_strs: T.List[str] = []

    appendix_properties: T.Set[str] = set()

    for prop in object_type.properties:
        conversion_type = edgedb_conversion_type_from_prop(prop)
        node_edgedb_conversion_map[prop.name] = FieldInfo(
            cast=conversion_type,
            cardinality=prop.cardinality,
            readonly=prop.readonly,
            required=prop.required,
        )
        if prop.is_computed:
            computed_properties.add(prop.name)
        is_appendix = False
        if node_config:
            if prop.name in node_config.appendix_properties:
                is_appendix = True
                appendix_properties.add(prop.name)
        if not prop.readonly and not prop.is_computed:
            updatable_fields.add(prop.name)
        if prop.is_exclusive:
            exclusive_fields.add(prop.name)
        default_value_str = "..." if prop.required else "None"
        # allow_mutation_str = (
        #     f"allow_mutation={not prop.readonly and not prop.is_computed}"
        # )
        if node_config and prop.name in node_config.basemodel_properties:
            prop_config = node_config.basemodel_properties[prop.name]
            module_name = prop_config.module_name
            type_str = prop.type_str_basemodel(
                module_name, cardinality=prop_config.cardinality
            )
        elif node_config and prop.name in node_config.custom_annotations:
            type_str = prop.type_str.replace(
                "str", node_config.custom_annotations[prop.name]
            )
        else:
            type_str = prop.type_str
        if not (prop.is_computed or is_appendix):
            property_strs.append(
                f"{prop.name}: {type_str} = Field({default_value_str})"
            )
        else:
            if prop.is_computed:
                exception_name = "ComputedPropertyException"
                # property_name = f"_{prop.name}"
                # property_str = f"{property_name}: T.Union[{type_str}, UnsetType] = PrivateAttr(UNSET)"
                property_name = f"{prop.name}_"
                property_str = f'{property_name}: T.Union[{type_str}, UnsetType] = Field(UNSET, alias="{prop.name}")'
            else:
                exception_name = "AppendixPropertyException"
                property_name = f"{prop.name}_"
                property_str = f'{property_name}: T.Union[{type_str}, UnsetType] = Field(UNSET, alias="{prop.name}")'
            exception_name = f"errors.{exception_name}"
            property_strs.append(property_str)
            val_str = "val"
            if "Set[" in type_str:
                val_str = f"val if type(val) != list else set(val)"
            computed_property_getter_strs.append(
                f"""
@property
def {prop.name}(self) -> {type_str}:
    # if self.{property_name} is UNSET:
    if "{property_name}" not in self.__fields_set__:
            raise {exception_name}("{prop.name} is unset")
    return self.{property_name} # type: ignore
                """
            )
        #             if not prop.is_computed:
        #                 computed_property_getter_strs.append(
        #                     f"""
        # @{prop.name}.setter
        # def {prop.name}(self, {prop.name}: {type_str}) -> None:
        #     self.{property_name} = {prop.name}
        #                     """
        #                 )
        if prop.name != "id" or allow_inserting_id:
            # for insert type
            if not prop.is_computed and not prop.not_insertable:
                insert_edgedb_conversion_map[prop.name] = FieldInfo(
                    cast=conversion_type,
                    cardinality=prop.cardinality,
                    readonly=prop.readonly,
                    required=prop.required,
                )
                insert_type_str = type_str
                # if required but has default, add optional back
                if prop.required and prop.default:
                    # insert_type_str = f"T.Optional[{insert_type_str}]"
                    insert_type_str = f"T.Union[{insert_type_str}, UnsetType]"
                if insert_type_str.startswith("T.Optional["):
                    insert_type_str_test = (
                        "T.Union[" + insert_type_str[11:-1] + ", None, UnsetType]"
                    )
                    insert_type_str = insert_type_str_test
                default_value_str = (
                    " = Field(UNSET)" if insert_type_str.endswith("UnsetType]") else ""
                )
                insert_property_strs.append(
                    f"{prop.name}: {insert_type_str}{default_value_str}"
                )
            # for update type
            if not prop.is_computed and not prop.readonly:
                patch_edgedb_conversion_map[prop.name] = FieldInfo(
                    cast=conversion_type,
                    cardinality=prop.cardinality,
                    readonly=prop.readonly,
                    required=prop.required,
                )
                patch_type_str = type_str
                if prop.required and prop.default:
                    patch_type_str = f"T.Optional[{patch_type_str}]"
                final_patch_type_str = f"T.Union[{patch_type_str}, UnsetType]"
                patch_property_strs.append(
                    f"{prop.name}: {final_patch_type_str} = Field(UNSET)"
                )

    link_function_strs: T.List[str] = []
    resolver_function_strs: T.List[str] = []
    updatable_links: T.Set[str] = set()
    exclusive_links: T.Set[str] = set()
    link_conversion_map: CONVERSION_MAP = {}
    edge_resolver_map: T.Dict[str, str] = {}

    for link in object_type.links:
        if link.name == "__type__":
            continue
        link_conversion_map[link.name] = FieldInfo(
            cast=link.target.model_name,
            cardinality=link.cardinality,
            readonly=link.readonly,
            required=link.required,
        )
        edge_resolver_map[link.name] = f"{link.target.model_name}Resolver"
        if not link.readonly and not link.is_computed:
            updatable_links.add(link.name)
        if link.is_exclusive:
            exclusive_links.add(link.name)
        link_function_strs.append(build_node_link_function_str(link))
        resolver_function_strs.append(
            build_resolver_link_function_str(
                node_resolver_name=node_resolver_name, link=link
            )
        )
        # for insert
        if not link.is_computed and not link.not_insertable:
            insert_resolver_str = f"{link.target.model_name}Resolver"
            if (not link.required) or (link.required and link.default):
                insert_resolver_str = f"T.Optional[{insert_resolver_str}]"
            default_value_str = (
                " = None" if insert_resolver_str.startswith("T.Optional[") else ""
            )
            insert_property_strs.append(
                f"{link.name}: {insert_resolver_str}{default_value_str}"
            )
        # for update
        if not link.is_computed and not link.readonly:
            patch_resolver_str = f"{link.target.model_name}Resolver"
            if (not link.required) or (link.required and link.default):
                patch_resolver_str = f"T.Optional[{patch_resolver_str}]"
            final_patch_resolver_str = f"T.Union[{patch_resolver_str}, UnsetType]"
            patch_property_strs.append(
                f"{link.name}: {final_patch_resolver_str} = Field(default_factory=UnsetType)"
            )

    basemodel_properties = (
        [] if not node_config else node_config.basemodel_properties.keys()
    )
    custom_annotations = (
        [] if not node_config else node_config.custom_annotations.keys()
    )
    mutate_on_update = {} if not node_config else node_config.mutate_on_update

    orm_config_str = build_orm_config(
        model_name=object_type.node_name,
        updatable_fields={*updatable_fields, *updatable_links},
        exclusive_fields={*exclusive_fields, *exclusive_links},
        appendix_properties=appendix_properties,
        computed_properties=computed_properties,
        basemodel_properties=basemodel_properties,
        custom_annotations=custom_annotations,
        mutate_on_update=mutate_on_update,
        node_edgedb_conversion_map=node_edgedb_conversion_map,
        insert_edgedb_conversion_map=insert_edgedb_conversion_map,
        patch_edgedb_conversion_map=patch_edgedb_conversion_map,
        insert_link_conversion_map=link_conversion_map,
    )

    insert_model_name = f"{object_type.node_name}Insert"
    patch_model_name = f"{object_type.node_name}Patch"

    # insert type
    insert_inner_str = "\n".join(insert_property_strs)
    insert_conversion_map_str = f"_edgedb_conversion_map: T.Dict[str, T.Dict[str, T.Union[str, bool]]] = {stringify_dict(insert_edgedb_conversion_map)}"
    # insert_s = f"class {insert_model_name}(Insert):\n{indent_lines(insert_inner_str)}\n\n{indent_lines(insert_conversion_map_str)}"
    insert_s = f"class {insert_model_name}(Insert):\n{indent_lines(insert_inner_str)}"

    # patch type
    patch_inner_str = "\n".join(patch_property_strs)
    patch_conversion_map_str = f"_edgedb_conversion_map: T.Dict[str, T.Dict[str, T.Union[str, bool]]] = {stringify_dict(patch_edgedb_conversion_map)}"
    # patch_s = f"class {patch_model_name}(Patch):\n{indent_lines(patch_inner_str)}\n\n{indent_lines(patch_conversion_map_str)}"
    patch_s = f"class {patch_model_name}(Patch):\n{indent_lines(patch_inner_str)}"

    # node
    node_properties_str = "\n".join(property_strs)
    from_str_validator_str = build_from_str_validator_str(node_config=node_config)
    if hydrate:
        node_properties_str = ""
        from_str_validator_str = ""

    computed_property_getter_str = "\n".join(computed_property_getter_strs)
    node_conversion_map_str = f"_edgedb_conversion_map: T.Dict[str, T.Dict[str, T.Union[str, bool]]] = {stringify_dict(node_edgedb_conversion_map)}"
    insert_link_conversion_map_str = f"_link_conversion_map: T.Dict[str, T.Dict[str, T.Union[str, bool]]] = {stringify_dict(link_conversion_map)}"
    computed_properties_str = f"_computed_properties: T.ClassVar[T.Set[str]] = {stringify_set(computed_properties)}"
    appendix_properties_str = f"_appendix_properties: T.ClassVar[T.Set[str]] = {stringify_set(appendix_properties)}"

    basemodel_properties_str = f"_basemodel_properties: T.ClassVar[T.Set[str]] = {stringify_set(set(basemodel_properties))}"

    custom_annotations_str = f"_custom_annotations: T.ClassVar[T.Set[str]] = {stringify_set(set(custom_annotations))}"
    node_link_functions_str = "\n".join(link_function_strs)

    if dehydrate:
        node_link_functions_str = ""
    node_inner_strs = [
        node_properties_str,
        "\n",
        from_str_validator_str,
        "\n",
        computed_property_getter_str,
        # node_conversion_map_str,
        # insert_link_conversion_map_str,
        # computed_properties_str,
        # appendix_properties_str,
        # basemodel_properties_str,
        # custom_annotations_str,
        node_link_functions_str,
        orm_config_str,
    ]
    node_inner_str = "\n".join(remove_falsies(node_inner_strs))
    inherits = f"Node" if not hydrate else f"{object_type.node_name}Hydrated"
    node_s = (
        f"class {object_type.node_name}({inherits}):\n{indent_lines(node_inner_str)}"
    )

    # resolver
    clses = [
        f"_node_cls = {object_type.node_name}",
        f"_insert_cls = {object_type.node_name}Insert",
        f"_patch_cls = {object_type.node_name}Patch",
    ]
    resolver_properties_str = "\n".join(clses)
    resolver_link_functions_str = "\n".join(resolver_function_strs)
    resolver_get_functions_str = build_exclusive_functions_str(
        node_name=object_type.node_name, exclusive_field_names=exclusive_fields
    )
    resolver_include_fields_str = build_include_fields_function(
        node_resolver_name=node_resolver_name,
        appendix_properties=appendix_properties,
        computed_properties=computed_properties,
    )
    resolver_filter_functions_str = build_filter_functions_str(
        node_name=object_type.node_name, conversion_map=node_edgedb_conversion_map
    )

    edge_resolver_val_strs = [
        f"T.Type[{v}]" for v in sorted(set(edge_resolver_map.values()))
    ]

    union_type = (
        f"T.Union[{', '.join(edge_resolver_val_strs)}]"
        if edge_resolver_map
        else "T.Type[Resolver]"
    )
    edge_resolver_map_strs.append(
        f"{node_resolver_name}._edge_resolver_map: T.Dict[str, {union_type}] ="
        f" {stringify_dict(edge_resolver_map, stringify_value=False)}"
    )

    resolver_inner_strs = [
        resolver_properties_str,
        resolver_link_functions_str,
        resolver_get_functions_str,
        resolver_filter_functions_str,
        resolver_include_fields_str,
    ]
    resolver_inner_str = "\n".join(resolver_inner_strs)
    resolver_mixin_str = "" if not resolver_mixin_model else f", {resolver_mixin_model}"
    resolver_s = f"class {node_resolver_name}(Resolver[{object_type.node_name}, {insert_model_name}, {patch_model_name}]{resolver_mixin_str}):\n{indent_lines(resolver_inner_str)}"

    return f"{node_s}\n{insert_s}\n{patch_s}\n{resolver_s}"


async def build_nodes_and_resolvers(
    client: edgedb.AsyncIOClient,
    db_config: DBConfig,
    nodes_to_hydrate: T.Set[str],
    dehydrate: bool,
) -> str:
    object_types = await introspect_objects(client)
    node_strs: T.List[str] = []
    edge_resolver_map_strs: T.List[str] = []
    resolver_mixin_model = (
        None
        if not db_config.resolver_mixin_path
        else db_config.resolver_mixin_path.split(" ")[-1]
    )
    for object_type in object_types:
        node_strs.append(
            build_node_and_resolver(
                object_type,
                node_config=db_config.nodes.get(object_type.node_name),
                edge_resolver_map_strs=edge_resolver_map_strs,
                hydrate=object_type.node_name in nodes_to_hydrate and not dehydrate,
                dehydrate=dehydrate,
                resolver_mixin_model=resolver_mixin_model,
            )
        )
    update_forward_refs_inserts_str = "\n".join(
        [f"{o.node_name}Insert.update_forward_refs()" for o in object_types]
    )
    update_forward_refs_patches_str = "\n".join(
        [f"{o.node_name}Patch.update_forward_refs()" for o in object_types]
    )
    update_forward_refs_nodes_str = "\n".join(
        [f"{o.node_name}.update_forward_refs()" for o in object_types]
    )
    nodes_str = "\n".join(node_strs)
    edge_resolver_map_str = "\n".join(edge_resolver_map_strs)
    return f"{nodes_str}\n\n{update_forward_refs_inserts_str}\n\n{update_forward_refs_patches_str}\n\n{update_forward_refs_nodes_str}\n\n{edge_resolver_map_str}"


def add_quotes_to_non_env_vars(s: str) -> str:
    if re.fullmatch(ENV_VAR_PATTERN, s) is not None:
        return s
    return f'"{s}"'


def build_client(db_config: DBConfig) -> str:
    return f"CLIENT = create_async_client(dsn={db_config.dsn_str()})"


def validate_output_path(path: Path) -> None:
    if not os.path.isdir(path):
        if os.path.isfile(path):
            raise GeneratorException(
                f"output path {path=} must be a directory, not a file."
            )
        if not os.path.exists(path):
            os.makedirs(path)


def build_hydrate_imports(db_config: DBConfig) -> str:
    import_strs: T.List[str] = []
    for node_name, config in db_config.nodes.items():
        if config.module_path:
            hydrated_name = f"{node_name}Hydrated"
            import_strs.append(
                f"from {config.module_path} import {node_name} as {hydrated_name}"
            )
            # idk why i have the part below, was just copying from dgraph_orm, seems unecessary and wrong
            # import_strs.append(
            #     f"{hydrated_name}.{CONFIG_NAME}.resolver._node = {hydrated_name}"
            # )
    return "\n".join(import_strs)


def get_nodes_to_hydrate(db_config: DBConfig) -> T.Set[str]:
    node_names: T.Set[str] = set()
    for node_name, config in db_config.nodes.items():
        if config.module_path:
            node_names.add(node_name)
    return node_names


async def build_enums_from_config(db_config: DBConfig, include_strawberry: bool) -> str:
    client = edgedb.create_async_client(dsn=db_config.dsn)
    enums_imports = build_enum_imports()
    enums_str = await build_enums(client, include_strawberry=include_strawberry)
    s = "\n".join([enums_imports, enums_str])
    if s.strip().endswith(":"):
        s += " pass"
    s = format_str(s, mode=FileMode())
    return s


async def build_from_config(
    db_config: DBConfig,
    enums_module: str,
    hydrate: bool = False,
    dehydrate: bool = False,
) -> str:
    client = edgedb.create_async_client(dsn=db_config.dsn)
    imports_str = imports(
        enums_module=enums_module, resolver_mixin_path=db_config.resolver_mixin_path
    )
    client_str = build_client(db_config)
    cache_only_str = f"CACHE_ONLY: bool = {db_config.cache_only}"
    validator_module_imports = build_validator_module_imports(db_config)
    hydrate_imports = "" if not hydrate else build_hydrate_imports(db_config)
    nodes_and_resolvers_str = await build_nodes_and_resolvers(
        client,
        db_config=db_config,
        nodes_to_hydrate=get_nodes_to_hydrate(db_config),
        dehydrate=dehydrate,
    )

    s = "\n".join(
        [
            imports_str,
            client_str,
            cache_only_str,
            validator_module_imports,
            hydrate_imports,
            nodes_and_resolvers_str,
        ]
    )
    s = format_str(s, mode=FileMode())
    return s


async def generate(
    *,
    config_path: Path = None,
    db_config_map: T.Dict[str, DBConfig] = None,
    output_path: Path,
    include_strawberry: bool = False,
) -> None:
    if config_path is not None and db_config_map is not None:
        raise GeneratorException(
            "Provide either a config_path or db_config_map, not both."
        )
    if config_path is not None:
        validate_output_path(output_path)
        all_config_json_str = open(config_path).read()
        all_config_d = parse_raw_as(T.Dict[str, DBConfig], all_config_json_str)
    elif db_config_map is not None:
        all_config_d = db_config_map
    else:
        raise GeneratorException("Provide either a config_path or a config_model_map.")
    for db_name, db_config in all_config_d.items():
        if db_config.copy_config:
            db_config_to_copy = all_config_d[db_config.copy_config]
            db_config.hydrate = db_config_to_copy.hydrate
            db_config.nodes = db_config_to_copy.nodes

        # first build enums folder
        enums_s = await build_enums_from_config(
            db_config=db_config, include_strawberry=include_strawberry
        )
        enums_module = f"{db_name}_enums"
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        open(output_path / f"{enums_module}.py", "w").write(enums_s)
        hydrate = db_config.hydrate
        # must include strawberry types in both for circular dependency reasons
        s = await build_from_config(
            db_config=db_config,
            dehydrate=hydrate,
            enums_module=enums_module,
        )
        open(output_path / f"{db_name}.py", "w").write(s)
        if hydrate:
            s = await build_from_config(
                db_config=db_config,
                hydrate=True,
                enums_module=enums_module,
            )
            open(output_path / f"{db_name}_hydrated.py", "w").write(s)



import typing as T
import orjson
from enum import Enum
from pydantic import BaseModel
from edge_orm.external import encoders
from edge_orm.node.models import Insert, Patch, CONVERSION_MAP, FieldInfo
from edge_orm.resolver import errors, enums

if T.TYPE_CHECKING:
    from .model import VARS
    from .model import Resolver

ResolverType = T.TypeVar("ResolverType", bound="Resolver")


def line_var_from_field_info(
    field_name: str, val: T.Any, field_info: FieldInfo, json_get_item: str = None
) -> tuple[str, T.Any]:
    type_cast = field_info.cast
    var_field_name = (
        f"${field_name}"
        if not json_get_item
        else f'json_get({json_get_item}, "{field_name}")'
    )
    field_str = f"{field_name} := <{type_cast}>{var_field_name}"
    if isinstance(val, (dict, list)):
        if type_cast.endswith("::str") or type_cast.endswith("::json"):
            val = orjson.dumps(encoders.jsonable_encoder(val))
    elif isinstance(val, BaseModel):
        val = val.json()
    elif isinstance(val, set):
        val = list(val)
        field_str = (
            f"{field_name} := array_unpack(<array<{type_cast}>>{var_field_name})"
        )
    elif isinstance(val, Enum):
        val = val.value
    elif val is None and not json_get_item:
        field_str = f"{field_name} := {{}}"
    return field_str, val


def line_var_from_resolver(
    model: Insert | Patch, field_name: str, r: ResolverType
) -> tuple[str, T.Any]:
    if isinstance(model, Patch):
        if r.update_operation is None:
            raise errors.ResolverException(
                "Link resolvers for Patch objects must have an update operation."
            )
    elif isinstance(model, Insert):
        if r.update_operation is None:
            r.update_operation = enums.UpdateOperation.REPLACE
        if r.update_operation != enums.UpdateOperation.REPLACE:
            errors.ResolverException(
                "Update operations for link resolvers for Insert objects cannot be anything "
                "other than None or REPLACE."
            )
    else:
        errors.ResolverException("Model must be either an Insert or Patch.")
    rez_s, rez_vars = r.full_query_str_and_vars(
        prefix=field_name, include_select=True, include_detached=True
    )
    s = f"{field_name} {r.update_operation.value} ({rez_s})"
    return s, rez_vars


def model_to_set_str_vars(
    *,
    model: Insert | Patch,
    conversion_map: CONVERSION_MAP,
    json_get_item: str = None,
    additional_link_str: str | None = None,
) -> tuple[str, "VARS"]:
    """takes in a model dictionary and returns a string that represents a mutation with this dictionary
    eg: {"name": "Jeremy Berman", "age": UNSET, "last_updated": 2022...} -> { name := <str>$name, age := <int>{}, ...}"""
    str_lst: list[str] = []
    variables: VARS = {}
    for field_name in model.__fields_set__:
        original_val = getattr(model, field_name)
        if field_name not in conversion_map:
            # this is a resolver
            rez_s, rez_vars = line_var_from_resolver(
                model=model, field_name=field_name, r=original_val
            )
            str_lst.append(rez_s)
            variables.update(rez_vars)
        else:
            field_str, val = line_var_from_field_info(
                field_name=field_name,
                val=original_val,
                field_info=conversion_map[field_name],
                json_get_item=json_get_item,
            )
            str_lst.append(field_str)
            if val is not None:
                variables[field_name] = val

    if additional_link_str:
        str_lst.append(additional_link_str)
    s = f'{{ {", ".join(str_lst)} }}'
    return s, variables



import typing as T
from pydantic import BaseModel, parse_raw_as
from pydantic.fields import ModelField

ModelType = T.TypeVar("ModelType", bound=BaseModel)


def from_str(
    cls: T.Type[BaseModel], v: str | ModelType, field: ModelField
) -> ModelType | None:
    if isinstance(v, str):
        if v == "null":
            return None
        res = parse_raw_as(field.outer_type_, v)
        return res
    else:
        return v


from enum import Enum

EnumType = T.TypeVar("EnumType", bound=Enum)


def transform_enum(
    original_value: str | Enum,
    new_enum_type: T.Type[EnumType],
    ignore_null: bool = False,
) -> T.Optional[EnumType]:
    if ignore_null:
        if original_value is None:
            return None
    if isinstance(original_value, Enum):
        og_val = original_value.value
    else:
        og_val = original_value
    new_value = (
        og_val.replace("/", "").replace("-", "_").replace("  ", " ").replace(" ", "_")
    )
    return new_enum_type(new_value)


def enum_from_str(
    cls: T.Type[BaseModel], v: EnumType | str, field: ModelField
) -> EnumType | None:
    if isinstance(v, str) or not isinstance(v, field.type_):
        return transform_enum(original_value=v, new_enum_type=field.type_)
    else:
        return v



from .logs import create_logger, logger
from .unset import UNSET, UnsetType
from .node import Node, NodeException, Insert, Patch, EdgeConfigBase
from .resolver import Resolver, ResolverException
from .resolver import enums as resolver_enums
from . import types_generator, validators
from .execute import ExecuteConstraintViolationException, ExecuteException

__all__ = [
    "UNSET",
    "UnsetType",
    "Node",
    "Insert",
    "Patch",
    "EdgeConfigBase",
    "Resolver",
    "NodeException",
    "ResolverException",
    "resolver_enums",
    "create_logger",
    "logger",
    "types_generator",
    "validators",
    "ExecuteConstraintViolationException",
    "ExecuteException",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['edge_orm',
 'edge_orm.external',
 'edge_orm.node',
 'edge_orm.resolver',
 'edge_orm.types_generator']

package_data = \
{'': ['*']}

install_requires = \
['black',
 'devtools',
 'edgedb>=1.0.0,<2.0.0',
 'orjson>=3.8.0,<4.0.0',
 'pydantic[email]>=1.10.2,<2.0.0']

extras_require = \
{'docs': ['mkdocs-material>=8.5.7,<9.0.0']}

setup_kwargs = {
    'name': 'edge-orm',
    'version': '0.1.7',
    'description': '',
    'long_description': '# edge orm',
    'author': 'Jeremy Berman',
    'author_email': 'jerber@sas.upenn.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import typing as T
from uuid import UUID
from enum import Enum
import edgedb
from pydantic import BaseModel, PrivateAttr
from edgedb import AsyncIOClient
from edge_orm.cache import Cache
from edge_orm.unset import UNSET
from .errors import NodeException

if T.TYPE_CHECKING:
    # from edge_orm.cache import Cache
    from edge_orm.resolver.model import Resolver


class PropertyCardinality(str, Enum):
    ONE = "ONE"
    MANY = "MANY"


class Cardinality(str, Enum):
    One = "One"
    Many = "Many"


class FieldInfo(BaseModel):
    cast: str
    cardinality: Cardinality
    readonly: bool
    required: bool


CONVERSION_MAP = dict[str, FieldInfo]


class EdgeConfigBase(BaseModel):
    model_name: str
    client: AsyncIOClient

    updatable_fields: set[str]
    exclusive_fields: set[str]

    appendix_properties: set[str]
    computed_properties: set[str]
    basemodel_properties: set[str]
    custom_annotations: set[str]
    mutate_on_update: dict[str, str]

    node_edgedb_conversion_map: CONVERSION_MAP
    insert_edgedb_conversion_map: CONVERSION_MAP
    patch_edgedb_conversion_map: CONVERSION_MAP

    insert_link_conversion_map: CONVERSION_MAP

    class Config:
        arbitrary_types_allowed = True


COMPUTED = dict[str, T.Any]


class Insert(BaseModel):
    pass

    class Config:
        allow_mutation = True
        validate_assignment = True
        arbitrary_types_allowed = True


class Patch(BaseModel):
    pass

    class Config:
        allow_mutation = True
        validate_assignment = True
        arbitrary_types_allowed = True


class classproperty(property):
    def __get__(self, owner_self, owner_cls):  # type: ignore
        return self.fget(owner_cls)  # type: ignore


class Node(BaseModel):
    id: UUID

    EdgeConfig: T.ClassVar[EdgeConfigBase]

    _computed: COMPUTED = PrivateAttr(default=dict())

    _cache: "Cache" = PrivateAttr(default_factory=Cache)

    _used_resolver: "Resolver" = PrivateAttr(None)  # type: ignore

    @property
    def computed(self) -> COMPUTED:
        return self._computed

    async def resolve(
        self,
        *,
        edge_name: str,
        edge_resolver: "Resolver",
        cache_only: bool = True,
        client: edgedb.AsyncIOClient | None = None,
    ) -> T.Any:
        val = self._cache.val_or_unset(edge=edge_name, resolver=edge_resolver)
        if val is UNSET:
            if cache_only:
                raise NodeException(
                    f"Could not get {edge_name} from the cache, and settings are cache_only."
                )
            else:
                new_r = self._used_resolver.__class__()
                # UserResolver().friends(edge_resolver)
                getattr(new_r, edge_name)(edge_resolver)
                this_node = await new_r._gerror(
                    field_name="id", value=self.id, client=client
                )
                new_val = await getattr(this_node, edge_name)(edge_resolver)
                self._cache.add(edge=edge_name, resolver=edge_resolver, val=new_val)
                return new_val
        return val

    """
    # TODO time the diff between initing it vs this property method
    @property
    def _cache(self) -> "Cache":
        if not self.__cache:
            self.__cache = Cache()
        return self.__cache
    """
    """
    @classproperty
    def Insert(self) -> T.Type[Insert]:  # example of how this could work
        return Insert
    """

    class Config:
        allow_mutation = False
        validate_assignment = True
        arbitrary_types_allowed = True



import typing as T
import time
from pydantic import BaseModel
from edge_orm.unset import UNSET, UnsetType

if T.TYPE_CHECKING:
    from edge_orm.node.models import Node
    from edge_orm.resolver.model import Resolver

RAW = dict[str, T.Any] | list[dict[str, T.Any]]


class CacheException(Exception):
    pass


class CacheNode(BaseModel):
    # val: T.Union["Node", list["Node"]]
    val: T.Any  # this could be a count or a string or a bool...
    # resolver: "Resolver"  # type: ignore
    resolver: T.Any  # unfortunately circular ^
    timestamp: float


class Cache(BaseModel):
    d: dict[str, list[CacheNode]] = {}

    def get(self, edge: str) -> list[CacheNode]:
        return self.d.get(edge, [])

    def has(self, edge: str) -> bool:
        return edge in self.d

    def add(self, edge: str, resolver: "Resolver", val: T.Any) -> None:  # type: ignore
        if not self.has(edge):
            self.d[edge] = []
        self.d[edge].append(
            CacheNode(resolver=resolver, val=val, timestamp=time.time())
        )

    def clear(self, edge: str) -> None:
        if edge in self.d:
            del self.d[edge]

    def is_empty(self) -> bool:
        return bool(self.d)

    def val(self, edge: str, resolver: "Resolver") -> T.Any:  # type: ignore
        for node in self.get(edge):
            if resolver.is_subset_of(node.resolver):
                return node.val
        raise CacheException(
            f"No node with edge {edge}, resolver {resolver.__dict__} found."
        )

    def val_or_unset(self, edge: str, resolver: "Resolver") -> T.Any:  # type: ignore
        try:
            return self.val(edge=edge, resolver=resolver)
        except CacheException:
            return UNSET
